import lief
import logging

from pathlib import Path
from typing import NewType, Dict, List

__all__ = [
    "Path",
    "LiefBin",
    "LiefSect",
    "FuncsInfo",
    "Unimplemented",
    "NotFound",
    "CompileFail",
    "Wrong",
    "logging",
]

LiefBin = NewType("liefBin", lief.ELF.Binary | lief.MachO.Binary | lief.PE.Binary)
LiefSect = NewType("liefSec", lief.ELF.Section | lief.MachO.Section | lief.PE.Section)
FuncsInfo = NewType("funcsInfo", List[Dict[list, int]])


class Unimplemented(Exception):
    def __init__(self, msg):
        pass


class NotFound(Exception):
    def __init__(self, msg):
        pass


class CompileFail(Exception):
    def __init__(self, cmd):
        pass


class Wrong(Exception):
    def __init__(self, msg):
        pass


logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.INFO)



#!/usr/bin/python3
import lief
import xml.etree.ElementTree as ET

from argparse import ArgumentParser

from internal.asm import Assembler
from internal.ftb import FuncTable
from internal.inj import Inject
from internal.cmpl import Compile
from shooker.internal.other import *


def main():
    parser = ArgumentParser()
    parser.add_argument(
        "--xml",
        type=Path,
        dest="config",
        default="hooks.xml",
        help="Hook config file (default: target_dir/hooks.xml)",
    )
    parser.add_argument("target_dir", type=Path, help="Directory with libs to hook")
    parser.add_argument("output_dir", type=Path, help="Directory to save hooked libs")
    args = parser.parse_args()

    args.output_dir.mkdir(parents=True, exist_ok=True)

    if not args.target_dir.exists():
        raise NotFound("Target dir")

    if args.config.name == "hooks.xml":
        args.config = args.target_dir / "hooks.xml"
    if not args.config.exists():
        raise NotFound("Hook config")

    with open(args.config) as xml_file:
        parser = ET.parse(xml_file)

    shook = parser.getroot()
    cc = shook.find("compiler").text

    for lib in shook.iterfind("lib_hook"):
        lib_path = (args.target_dir / lib.attrib["path"]).name
        logging.info(f"Patching {lib_path}...")

        target = lief.parse(lib_path)

        cs_arch = lib.find("arch")
        cs_mode = lib.find("mode")
        if cs_arch is None or cs_mode is None:
            raise NotFound("Arch or mode for hooked liblary")
        asm = Assembler(target, cs_arch.text, cs_mode.text)
        ftb = FuncTable(target, asm)
        cmpl = Compile(cc, args.target_dir)
        inj = Inject(target, asm)

        # parse included libs
        for lib in lib.find("include").iterfind("lib"):
            kind = inc.attrib.get("kind")
            if kind == "system":
                inc_value = f"#include <{inc}>"
            elif kind == "local":
                inc_value = f'#include "{inc}"'
            else:
                raise Wrong("Kind of include lib")
            cmpl.include_lib(inc_value)

        # parse included funcs
        for inc in lib.find("include").iterfind("func"):
            kind = inc.attrib.get("kind")
            name = inc.text
            if kind == "import":
                addr = ftb.load_import(name)
            elif kind == "symbol":
                addr = ftb.load_symbol(name)
            else:
                raise Wrong("Kind of include func")
            proto = inc.attrib.get("proto")
            cmpl.include_func(name, proto, addr)

        # define included libs&funcs
        cmpl.assemble_transl()

        # compile hooks
        for to_hook in lib.iterfind("hook"):
            fnc_name = to_hook.attrib.get("name")
            fnc_proto = to_hook.attrib.get("proto")
            fnc_code = to_hook.text

            logging.info(f"Compiling hook for {fnc_name}")
            cmpl.add_func_to_transl(fnc_name, fnc_proto, fnc_code)
        segm_addr = target.get_section(".shook").virtual_address
        funcs_info = cmpl.compile_transl(segm_addr)

        logging.info("Patching the hook(s)...")
        content = []
        for func in funcs_info.values():
            arr = func["content"]
            offset = func["offset"] + segm_addr
            content += asm.patch_sub_values(arr, offset)
        inj.shook_sect_fill(content)

        for to_hook in lib.iterfind("hook"):
            fnc_name = to_hook.attrib["name"]
            logging.info(f"Hooking {fnc_name}")

            fnc_offset = target.get_static_symbol(fnc_name).value  # address of func
            # offset of func in created section
            payl_offset = funcs_info[fnc_name]["offset"]
            inj.hook(fnc_name, fnc_offset, payl_offset)

        target.write((args.output_dir / lib.attrib["path"]).name)
    logging.info("Lib(s) patched")


if __name__ == "__main__":
    main()



import capstone

from lief.ELF import ARCH

from shooker.internal.other import *


class Assembler:
    """Arch-depended stuff"""

    def __init__(self, target: LiefBin, arch: str, mode: str) -> None:
        self.bin = target
        self.arch = self.bin.header.machine_type
        cs_arch, cs_mode = getattr(capstone, arch), getattr(capstone, mode)
        self.md = capstone.Cs(cs_arch, cs_mode)

    def jump(self, target_addr: int) -> bytes:
        """Create jmp-instruction to target_addr"""
        if self.arch == ARCH.x86_64:
            target_addr -= 5  # length of jmp instr
            jump_inst = b"\xE9" + target_addr.to_bytes(4, "little")
        else:
            raise Unimplemented(f"Assemble jump for {arch}")

        logging.debug(
            f"Crafted jump-instruction is {jump_inst.hex()} for {hex(target_addr)}"
        )
        return jump_inst

    def brute_ptl(self, section: LiefSect, target_addr: int) -> int:
        """Find plt-stub for specific function from .got.plt"""
        cont = section.content.tobytes()
        plt_offs = section.virtual_address

        if self.arch == ARCH.x86_64:
            for instr in self.md.disasm(cont, plt_offs):
                if instr.mnemonic != "jmp" or "word ptr [rip" not in instr.op_str:
                    continue

                instr_offs = int.from_bytes(instr.bytes[2:], "little")

                # offset from instruction + current entry plt position + len of jmp instruction
                if instr_offs + instr.address + instr.size == target_addr:
                    logging.debug(f"Found plt for {hex(target_addr)}")
                    return instr.address
        else:
            raise Unimplemented(f"Enumerating plt for {self.arch}")

        raise NotFound("Include function not found")

    def patch_sub_values(self, func_cont: list, func_offs: int) -> list:
        """Patch values that substracted from rip for relative-call instruction"""

        cont = bytearray(b"".join([i.to_bytes(1, "big") for i in func_cont]))
        register = None
        lea_ip = 0
        sub_instr = None
        call_ip = 0

        if self.arch == ARCH.x86_64:
            for instr in self.md.disasm(cont, 0):
                # print(instr)
                if instr.mnemonic == "lea" and "[rip - 7]" in instr.op_str:
                    lea_ip = instr.address
                    register = instr.op_str.replace(", [rip - 7]", "")
                    logging.debug(
                        f"Register of lea instr found {register} at {hex(func_offs+lea_ip)}"
                    )

                if (
                    instr.mnemonic == "sub"
                    and register is not None
                    and instr.op_str.startswith(register)
                ):
                    sub_instr = instr
                    logging.debug(f"Sub instr found at {hex(func_offs+instr.address)}")

                if (
                    instr.mnemonic == "call"
                    and sub_instr is not None
                    and 2 <= instr.size <= 3
                ):
                    call_ip = instr.address
                    logging.debug(f"Call instr found at {hex(func_offs+call_ip)}")

                if lea_ip != 0 and sub_instr is not None and call_ip != 0:
                    sub_value = int.from_bytes(sub_instr.bytes[2:], "little")
                    relative_offs = func_offs + lea_ip - sub_value
                    sub_bytes = sub_instr.bytes[:2] + relative_offs.to_bytes(4, "little")

                    cont[sub_instr.address : sub_instr.address + sub_instr.size] = sub_bytes

                    # zeroing to continue parse function
                    lea_ip, call_ip = 0, 0
                    register, sub_instr = None, None

        else:
            raise Unimplemented(f"Patch sub-instruction for {arch}")

        return [i for i in cont]



from lief import ELF
from dataclasses import dataclass

from shooker.internal.other import *


class FuncTable:
    """To resolve funcs"""

    def __init__(self, target: LiefBin, asm) -> None:
        self.bin = target
        self.asm = asm
        self.fncs = []

    def load_symbol(self, fnc_name: str) -> int:
        """Get definition of symbol from .text section in hooked lib"""
        fnc = self.bin.get_static_symbol(fnc_name)
        if not isinstance(fnc, ELF.Symbol):
            raise NotFound(f"Static function {fnc_name}")

        entry = FuncEntry(fnc.name, fnc.value)
        self.fncs.append(entry)
        logging.debug(f"Static {entry.name} loaded at {hex(entry.addr)}")
        return fnc.value

    def load_import(self, fnc_name: str) -> int:
        """Get definition of import symbol from .plt section in hooked lib"""
        fnc = self.bin.get_relocation(fnc_name)
        if not isinstance(fnc, ELF.Relocation):
            raise NotFound(f"Dynamic function {fnc_name}")

        plt = self.bin.get_section(".plt")
        jump_addr = self.asm.brute_ptl(plt, fnc.address)
        entry = FuncEntry(fnc.symbol.name, jump_addr)
        self.fncs.append(entry)
        logging.debug(f"Import {entry.name} loaded at {hex(entry.addr)}")
        return jump_addr


@dataclass
class FuncEntry:
    name: str
    addr: int
